import pytest

from excitertools import Iter


def test_chunked():
    it = Iter(range(12))
    assert it.chunked(3).collect() == [
        [0, 1, 2],
        [3, 4, 5],
        [6, 7, 8],
        [9, 10, 11],
    ]


def test_ichunked():
    it = Iter(range(12))
    chunks = it.ichunked(3).collect()
    assert list(chunks[3]) == [9, 10, 11]
    assert list(chunks[2]) == [6, 7, 8]
    assert list(chunks[1]) == [3, 4, 5]
    assert list(chunks[0]) == [0, 1, 2]


def test_sliced():
    seq = list(range(4))
    assert Iter.sliced(seq, 2).collect() == [[0, 1], [2, 3]]
    assert Iter.sliced(seq, 3).collect() == [[0, 1, 2], [3]]


def test_distribute():
    g1, g2 = Iter.range(6).distribute(2)
    assert g1.collect() == [0, 2, 4]
    assert g2.collect() == [1, 3, 5]


def test_divide():
    g1, g2 = Iter.range(6).divide(2)
    assert g1.collect() == [0, 1, 2]
    assert g2.collect() == [3, 4, 5]


def test_split_at():
    it = Iter("abcdcba").split_at(lambda x: x == "b")
    assert it.collect() == [
        ["a"],
        ["c", "d", "c"],
        ["a"],
    ]


def test_split_before():
    it = Iter("OneTwo").split_before(lambda s: s.isupper())
    assert it.collect() == [
        ["O", "n", "e"],
        ["T", "w", "o"],
    ]


def test_split_after():
    it = Iter("one1two2").split_after(lambda s: s.isdigit())
    assert it.collect() == [
        ["o", "n", "e", "1"],
        ["t", "w", "o", "2"],
    ]


def test_bucket():
    iterable = ["a1", "b1", "c1", "a2", "b2", "c2", "b3"]
    bucket = Iter(iterable).bucket(key=lambda x: x[0])
    assert bucket["a"].collect() == ["a1", "a2"]
    assert bucket["b"].collect() == ["b1", "b2", "b3"]

    bucket = Iter(iterable).bucket(key=lambda x: x[0])
    a_iterable = bucket["a"]
    for x in a_iterable:
        print(x)

    bucket = Iter(iterable).bucket(key=lambda x: x[0])
    a_iterable = bucket["a"]
    assert next(a_iterable) == "a1"
    assert next(a_iterable) == "a2"

    bucket = Iter(iterable).bucket(key=lambda x: x[0])
    a_iterable = bucket["a"]
    result = a_iterable.collect()
    assert result == ["a1", "a2"]

    bucket = Iter(iterable).bucket(key=lambda x: x[0])
    result = Iter(bucket).collect()
    assert result == ["a", "b", "c"]

    bucket = Iter(iterable).bucket(key=lambda x: x[0])
    for k, expected in zip(bucket, ["a", "b", "c"]):
        assert k == expected

    result = Iter(iterable).bucket(key=lambda x: x[0])["a"].collect()
    assert result == ["a1", "a2"]


def test_unzip():
    iterable = [("a", 1), ("b", 2), ("c", 3), ("d", 4)]
    it = Iter(iterable).unzip()
    parts = it.collect()
    assert parts[0].collect() == ["a", "b", "c", "d"]
    assert parts[1].collect() == [1, 2, 3, 4]


def test_grouper():
    assert Iter("ABCDEFG").grouper(3, "x").collect() == [
        ("A", "B", "C"),
        ("D", "E", "F"),
        ("G", "x", "x"),
    ]


def test_partition():
    left, right = Iter(range(10)).partition(lambda x: x % 2 != 0)
    assert left.collect() == [0, 2, 4, 6, 8]
    assert right.collect() == [1, 3, 5, 7, 9]

    def odd(x):
        return x % 2 != 0

    left = Iter(range(10)).partition(odd).nth(0).collect()

    right = Iter(range(10)).partition(odd).nth(1).collect()
    assert left == [0, 2, 4, 6, 8]
    assert right == [1, 3, 5, 7, 9]


def test_spy():
    head, iterable = Iter("abcdefg").spy()
    assert head.collect() == ["a"]
    assert iterable.collect() == ["a", "b", "c", "d", "e", "f", "g"]


def test_peekable():
    p = Iter(["a", "b"]).peekable()
    assert p.peek() == "a"
    assert next(p) == "a"

    assert Iter([]).peekable().peek("hi") == "hi"

    p = Iter([1, 2, 3]).peekable()
    p.prepend(10, 11, 12)
    assert next(p) == 10
    assert p.peek() == 11
    # TODO: is there a clean way to keep p an Iter instance?
    assert Iter(p).collect() == [11, 12, 1, 2, 3]

    p = Iter([1, 2, 3]).peekable()
    p.prepend(10, 11, 12)
    assert p.peek() == 10
    assert p[2] == 12
    assert p[-1] == 3
    assert next(p) == 10
    assert p[2] == 1


def test_seekable():
    it = Iter.count().map(str).seekable()
    assert (next(it), next(it), next(it)) == ("0", "1", "2")

    it.seek(0)
    assert (next(it), next(it), next(it)) == ("0", "1", "2")

    assert next(it) == "3"

    it.seek(10)
    assert next(it) == "10"


def test_map_reduce_keyfunc():
    x = Iter("abbccc").map_reduce(lambda x: x.upper())
    assert sorted(x.items()) == [
        ("A", ["a"]),
        ("B", ["b", "b"]),
        ("C", ["c", "c", "c"]),
    ]


def test_map_reduce_valuefunc():
    x = Iter("abbccc").map_reduce(lambda x: x.upper(), lambda x: 1)
    assert sorted(x.items()) == [("A", [1]), ("B", [1, 1]), ("C", [1, 1, 1])]


def test_map_reduce_reducefunc():
    x = Iter("abbccc").map_reduce(
        lambda x: x.upper(),
        lambda x: 1,
        sum,
    )
    assert sorted(x.items()) == [("A", 1), ("B", 2), ("C", 3)]


def test_interleave():
    result = Iter("caleb").interleave(Iter.repeat("x")).collect()
    assert result == list("cxaxlxexbx")

    result = Iter("caleb").interleave(Iter.repeat("x")).concat("")
    assert result == "cxaxlxexbx"


def test_split_into():
    result = Iter("caleb").split_into([2, 1, 2]).collect()
    assert result == [["c", "a"], ["l"], ["e", "b"]]


def test_split_when():
    result = (
        Iter([1, 2, 3, 1, 4, 100, 4, 5, 1, 2]).split_when(lambda x, y: y < x).collect()
    )
    assert result == [[1, 2, 3], [1, 4, 100], [4, 5], [1, 2]]


@pytest.mark.parametrize(
    "iterable,n,fillvalue,step,expected",
    [
        ([1, 2, 3, 4, 5], 3, None, 1, [(1, 2, 3), (2, 3, 4), (3, 4, 5)]),
        ([1, 2, 3], 4, None, 1, [(1, 2, 3, None)]),
        ([1, 2, 3, 4, 5, 6], 3, "!", 2, [(1, 2, 3), (3, 4, 5), (5, 6, "!")]),
    ],
)
def test_windowed(iterable, n, fillvalue, step, expected):
    result = Iter(iterable).windowed(n, fillvalue, step).collect()
    assert result == expected


@pytest.mark.parametrize(
    "iterable,expected",
    [
        ("more", ["m", "o", "r", "e", "mo", "or", "re", "mor", "ore", "more"]),
    ],
)
def test_substrings(iterable, expected):
    result = Iter(iterable).substrings().map(lambda s: "".join(s)).collect()
    assert result == expected


@pytest.mark.parametrize(
    "iterable,expected",
    [
        (
            "more",
            [
                ("m", 0, 1),
                ("o", 1, 2),
                ("r", 2, 3),
                ("e", 3, 4),
                ("mo", 0, 2),
                ("or", 1, 3),
                ("re", 2, 4),
                ("mor", 0, 3),
                ("ore", 1, 4),
                ("more", 0, 4),
            ],
        ),
    ],
)
def test_substrings_indexes(iterable, expected):
    result = (
        Iter(iterable)
        .substrings_indexes()
        .map(lambda tup: ("".join(tup[0]), *tup[1:]))
        .collect()
    )
    assert result == expected



import excitertools
from collections import Counter
from excitertools import apply


def stem(word):
    """Stem word to primitive form

    This example taken from toolz: https://github.com/pytoolz/toolz

    There are several different ways to split a string, stem the words,
    and calculate their frequency.

    """
    return word.lower().rstrip(",.!:;'-\"").lstrip("'\"")


def test_stem():
    sentence = "This cat jumped over this other cat!"
    wc = excitertools.Iter(sentence.split()).map(stem).collect(Counter)
    assert wc == Counter({"this": 2, "cat": 2, "jumped": 1, "over": 1, "other": 1})


def test_stem2():
    sentence = "This cat jumped over this other cat!"
    wc = apply(str.split, sentence).map(stem).collect(Counter)
    assert wc == Counter({"this": 2, "cat": 2, "jumped": 1, "over": 1, "other": 1})


def test_stem3():
    sentence = "This cat jumped over this other cat!"
    wc = excitertools.splititer_regex(r"\s+", sentence).map(stem).collect(Counter)
    assert wc == Counter({"this": 2, "cat": 2, "jumped": 1, "over": 1, "other": 1})



import pytest

from excitertools import Iter, insert_separator, concat


def test_basic():
    it = Iter(range(5))
    assert list(it) == [0, 1, 2, 3, 4]


@pytest.mark.parametrize(
    "container,it,expected",
    [
        (list, range(5), [0, 1, 2, 3, 4]),
        (set, range(5), {0, 1, 2, 3, 4}),
        (dict, zip("abc", range(3)), dict(a=0, b=1, c=2)),
        (dict.fromkeys, "abc", dict(a=None, b=None, c=None)),
        (frozenset, range(5), frozenset([0, 1, 2, 3, 4])),
        (str, ["a", "b", "c"], "abc"),
        # (bytes, [b'a', b'b', b'c'], b'abc'),
    ],
)
def test_collect(container, it, expected):
    it = Iter(it)
    assert it.collect(container=container) == expected


def test_islice():
    it = Iter(range(5))
    assert it.islice(3).collect() == [0, 1, 2]


def test_cycle_islice():
    it = Iter(range(5))
    assert it.cycle().islice(6, 9, 2).collect() == [1, 3]


def test_count():
    it = Iter.count().islice(3, 6)
    assert it.collect() == [3, 4, 5]


def test_map():
    it = Iter.range(5).map(lambda x: x * 2)
    assert it.collect() == [0, 2, 4, 6, 8]

    it = Iter.range(5).map(lambda x: x * 2).filter(lambda x: x > 5)
    assert it.collect() == [6, 8]

    result = (
        Iter.range(5)
        .map(lambda x: x * 2)
        .filter(lambda x: x > 5)
        .reduce(lambda x, y: x + y)
    )
    assert result == 14


def test_piecemeal_results():
    it = Iter.range(5).map(lambda x: x * 2).filter(lambda x: x > 5)

    x = iter(it)
    assert next(x) == 6
    assert next(x) == 8


@pytest.mark.parametrize(
    "v,expected",
    [
        (8, True),
        (9, False),
    ],
)
def test_any(v, expected):
    result = (
        Iter.range(10)
        .map(lambda x: x * 2)
        .filter(lambda x: x > 5)
        .map(lambda x: x == v)
        .any()
    )
    assert result == expected


@pytest.mark.parametrize(
    "v,expected",
    [
        (5, True),
        (6, False),
    ],
)
def test_all(v, expected):
    assert (
        len(
            Iter.range(0, 100, 5)
            .filter(lambda x: x > 5 and x < 90)
            .map(lambda x: x % v == 0)
            .collect()
        )
        > 0
    )
    result = (
        Iter.range(0, 100, 5)
        .filter(lambda x: x > 5 and x < 90)
        .map(lambda x: x % v == 0)
        .all()
    )
    assert result == expected


def test_sum():
    result = Iter.range(5).sum()
    assert result == 10


@pytest.mark.parametrize(
    "input,glue,output",
    [
        ("abc", "-", "a-b-c"),
        (iter("abc"), "-", "a-b-c"),
        (b"abc", b"-", b"a-b-c"),
        (b"abc", b"--", b"a--b--c"),
        (bytearray(b"abc"), b"--", b"a--b--c"),
        (b"abc", bytearray(b"--"), b"a--b--c"),
    ],
)
def test_concat(input, glue, output):
    assert concat(input, glue) == output


def test_concat_error():
    with pytest.raises(ValueError):
        assert concat("caleb", None) == "c-a-l-e-b"


@pytest.mark.parametrize(
    "input,glue,output",
    [
        ([1, 2, 3], None, [1, 2, 3]),
        ([1, 2, 3], 0, [1, 0, 2, 0, 3]),
        (iter([1, 2, 3]), 0, [1, 0, 2, 0, 3]),
        ([1, 2, 3], [8, 9], [1, [8, 9], 2, [8, 9], 3]),
        ("abc", "-", list("a-b-c")),
        pytest.param(
            b"abc",
            b"-",
            list(b"a-b-c"),
            marks=pytest.mark.xfail(reason="bytes are dumb"),
        ),
        # (b'abc', b'--', list(b'a--b--c')),
        # (bytearray(b'abc'), b'--', list(b'a--b--c')),
        # (b'abc', bytearray(b'--'), list(b'a--b--c')),
    ],
)
def test_insert_separator(input, glue, output):
    assert list(insert_separator(input, glue)) == output


def test_insert():
    result = Iter("caleb").insert("x").collect(container=str)
    # Note the difference as compare to the output of interleave
    assert result == "cxaxlxexb"


@pytest.mark.parametrize(
    "elem,times,expected",
    [
        ("x", 3, ["x"] * 3),
    ],
)
def test_repeat_finite(elem, times, expected):
    result = Iter.repeat(elem, times=times).collect()
    assert result == expected


@pytest.mark.parametrize(
    "elem,taken,expected",
    [
        ("x", 0, []),
        ("x", 3, ["x"] * 3),
    ],
)
def test_repeat_infinite(elem, taken, expected):
    result = Iter.repeat(elem).take(taken).collect()
    assert result == expected


@pytest.mark.parametrize(
    "arg,func,expected",
    [
        ([1, 2, 3, 4, 5], None, [1, 3, 6, 10, 15]),
        ([1, 2, 3, 4, 5], lambda x, y: x * y, [1, 2, 6, 24, 120]),
    ],
)
def test_accumulate(arg, func, expected):
    result = Iter(arg).accumulate(func).collect()
    assert result == expected


@pytest.mark.parametrize(
    "arg,iterables,expected",
    [
        ("ABC", ["DEF"], ["A", "B", "C", "D", "E", "F"]),
    ],
)
def test_chain(arg, iterables, expected):
    result = Iter(arg).chain(*iterables).collect()
    assert result == expected


@pytest.mark.parametrize(
    "args,expected",
    [
        (["ABC", "DEF"], ["A", "B", "C", "D", "E", "F"]),
    ],
)
def test_from_iterable(args, expected):
    result = Iter(args).chain_from_iterable().collect()
    assert result == expected


@pytest.mark.parametrize(
    "arg,selectors,expected",
    [
        ("ABCDEF", [1, 0, 1, 0, 1, 1], ["A", "C", "E", "F"]),
    ],
)
def test_compress(arg, selectors, expected):
    result = Iter(arg).compress(selectors).collect()
    assert result == expected


@pytest.mark.parametrize(
    "arg,pred,expected",
    [
        ([1, 4, 6, 4, 1], lambda x: x < 5, [6, 4, 1]),
    ],
)
def test_dropwhile(arg, pred, expected):
    result = Iter(arg).dropwhile(pred).collect()
    assert result == expected


@pytest.mark.parametrize(
    "arg,pred,expected",
    [
        (range(10), lambda x: x % 2, [0, 2, 4, 6, 8]),
    ],
)
def test_filterfalse(arg, pred, expected):
    result = Iter(arg).filterfalse(pred).collect()
    assert result == expected


@pytest.mark.parametrize(
    "arg,key,expected",
    [
        ("AABCBA", None, [["A", "A", "A"], ["B", "B"], ["C"]]),
    ],
)
def test_groupby(arg, key, expected):
    # NOTE: this doesn't work because the underlying iterator is shared,
    # see the docs for more info. You have to do something with the groups
    # while the iterator is in-flight.
    result = Iter(sorted(arg)).groupby(key).map(lambda tup: list(tup[1])).collect()
    assert result == expected


@pytest.mark.parametrize(
    "arg,func,expected",
    [
        (
            [("a", 1), ("b", 2), ("c", 3)],
            lambda c, n: f"{c}={n}",
            ["a=1", "b=2", "c=3"],
        ),
    ],
)
def test_starmap(arg, func, expected):
    result = Iter(arg).starmap(func).collect()
    assert result == expected


def test_takewhile():
    result = Iter([1, 4, 6, 4, 1]).takewhile(lambda x: x < 5).collect()
    assert result == [1, 4]


def test_tee():
    x, y = Iter([1, 4, 6, 4, 1]).tee()
    assert x.take(2).collect() == [1, 4]
    assert y.take(3).collect() == [1, 4, 6]

    result = Iter([1, 4, 6, 4, 1]).tee(3).nth(2).take(2).collect()
    assert result == [1, 4]


def test_zip():
    result = Iter([1, 4, 6, 4, 1]).zip("abc").collect()
    assert result == [(1, "a"), (4, "b"), (6, "c")]


def test_zip_longest():
    result = Iter([1, 4, 6, 4, 1]).zip_longest("abc").collect()
    assert result == [(1, "a"), (4, "b"), (6, "c"), (4, None), (1, None)]


def test_non_iterable_error():
    """
    Test the error message we give if a user accidentally passes a
    generator function without evaluating it. This is only to give
    a friendlier error message for a fairly common error. I still do
    it sometimes so this helps me as much as anyone.
    """

    def gen():
        for x in range(5):
            yield x

    with pytest.raises(TypeError, match="seems you passed"):
        Iter(gen)

    with pytest.raises(TypeError, match="'object' object is not iterable"):
        Iter(object())



import pytest
from excitertools import Iter, IterDict


def test_basic():
    it = Iter(range(5))
    assert list(it) == [0, 1, 2, 3, 4]


@pytest.mark.parametrize(
    "val,wargs,jargs,result",
    [
        ("caleb", [], [], "(caleb)"),
        ("caleb", ['""'], [], '"caleb"'),
        ("caleb", [('"""', '"""')], [], '"""caleb"""'),
        ("caleb", [], [","], "(,c,a,l,e,b,)"),
    ],
)
def test_wrap(val, wargs, jargs, result):
    assert Iter(val).wrap(*wargs).concat(*(jargs or [""])) == result


@pytest.mark.parametrize(
    "val,wargs,jargs,result",
    [
        (iter("caleb"), [(iter("abc"), iter("def"))], [""], "abccalebdef"),
    ],
)
def test_wrap_iterators(val, wargs, jargs, result):
    assert Iter(val).wrap(*wargs).concat(*jargs) == result


@pytest.mark.parametrize(
    "val,wargs,result",
    [
        ([1, 2, 3], [([0, 0], [9, 9])], [0, 0, 1, 2, 3, 9, 9]),
    ],
)
def test_wrap_iterators2(val, wargs, result):
    assert Iter(val).wrap(*wargs).collect() == result


def test_wrap_error():
    with pytest.raises(ValueError):
        Iter("caleb").wrap(ends="x").join()


def test_enumerate():
    result = Iter("caleb").enumerate().collect()
    assert result == [
        (0, "c"),
        (1, "a"),
        (2, "l"),
        (3, "e"),
        (4, "b"),
    ]

    result = Iter("caleb").enumerate().filter(lambda x: x[1] in "ab").collect()
    assert result == [
        (1, "a"),
        (4, "b"),
    ]


def test_dict():
    result = Iter("caleb").enumerate().dict()
    assert result == {0: "c", 1: "a", 2: "l", 3: "e", 4: "b"}

    result = Iter("caleb").map(lambda x: (x, ord(x))).dict()
    assert result == {"a": 97, "b": 98, "c": 99, "e": 101, "l": 108}

    result = Iter("caleb").map("x, ord(x)").dict()
    assert result == {"a": 97, "b": 98, "c": 99, "e": 101, "l": 108}


def test_iterdict():
    result = IterDict().__iter__().collect()
    assert result == []

    result = IterDict(dict(a=1, b=2)).__iter__().collect()
    assert result == ["a", "b"]

    result = IterDict([("a", 1), ("b", 2)]).__iter__().collect()
    assert result == ["a", "b"]

    result = IterDict(a=1, b=2, c=3).__iter__().collect()
    assert result == ["a", "b", "c"]

    result = [x for x in IterDict(a=1, b=2, c=3)]
    assert result == ["a", "b", "c"]

    result = IterDict(a=1, b=2, c=3).keys().collect()
    assert result == ["a", "b", "c"]

    result = IterDict(a=1, b=2, c=3).values().collect()
    assert result == [1, 2, 3]

    result = IterDict(a=1, b=2, c=3).items().collect()
    assert result == [("a", 1), ("b", 2), ("c", 3)]

    result = IterDict(a=1, b=2, c=3).items().filter(lambda x: x[1] % 2).collect()
    assert result == [("a", 1), ("c", 3)]

    result = IterDict(a=1, b=2, c=3).items().filter(lambda x: x[0] == "b").collect()
    assert result == [("b", 2)]

    result = (
        IterDict(a=1, b=2, c=3)
        .update(d=4, e=5)
        .items()
        .filter(lambda x: x[1] % 2)
        .collect()
    )
    assert result == [("a", 1), ("c", 3), ("e", 5)]



import pathlib
from excitertools import Iter


def test_open():
    fname = pathlib.Path(__file__)
    x = Iter.open(fname).map(str.strip).filter(None).map(str.upper).collect()
    assert len(x) == 7
    assert x[0] == "IMPORT PATHLIB"
